@@ -58,15 +58,15 @@ struct timeval now;
 unsigned char myid[8];
 int debug = 0;
 
-time_t reboot_time;
-
 int link_detect = 0;
 int all_wireless = 0;
 int default_wireless_hello_interval = -1;
 int default_wired_hello_interval = -1;
 int resend_delay = -1;
 int do_daemonise = 0;
-char *logfile = NULL, *pidfile = "/var/run/babeld.pid";
+char *logfile = NULL,
+    *pidfile = "/var/run/babeld.pid",
+    *state_file = "/var/lib/babel-state";
 
 unsigned char *receive_buffer = NULL;
 int receive_buffer_size = 0;
@@ -76,8 +76,6 @@ const unsigned char ones[16] =
     {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
      0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
 
-char *state_file = "/var/lib/babel-state";
-
 int protocol_port;
 unsigned char protocol_group[16];
 int protocol_socket = -1;
@@ -86,13 +84,10 @@ static int kernel_routes_changed = 0;
 static int kernel_link_changed = 0;
 static int kernel_addr_changed = 0;
 
-struct timeval check_neighbours_timeout;
+struct timeval check_neighbours_timeout, check_interfaces_timeout;
 
 static volatile sig_atomic_t exiting = 0, dumping = 0, reopening = 0;
 
-int local_server_socket = -1, local_socket = -1;
-int local_server_port = -1;
-
 static int kernel_routes_callback(int changed, void *closure);
 static void init_signals(void);
 static void dump_tables(FILE *out);
@@ -102,7 +97,7 @@ int
 main(int argc, char **argv)
 {
     struct sockaddr_in6 sin6;
-    int rc, fd, i, opt;
+    int rc, fd, i, opt, random_id = 0;
     time_t expiry_time, source_expiry_time, kernel_dump_time;
     char *config_file = NULL;
     void *vrc;
@@ -122,9 +117,10 @@ main(int argc, char **argv)
 
     parse_address("ff02:0:0:0:0:0:1:6", protocol_group, NULL);
     protocol_port = 6696;
+    change_smoothing_half_life(4);
 
     while(1) {
-        opt = getopt(argc, argv, "m:p:h:H:i:k:A:PsuS:d:g:lwz:t:T:c:C:DL:I:");
+        opt = getopt(argc, argv, "m:p:h:H:i:k:A:sruS:d:g:lwz:M:t:T:c:C:DL:I:");
         if(opt < 0)
             break;
 
@@ -145,7 +141,9 @@ main(int argc, char **argv)
             }
             break;
         case 'p':
-            protocol_port = atoi(optarg);
+            protocol_port = parse_nat(optarg);
+            if(protocol_port <= 0 || protocol_port > 0xFFFF)
+                goto usage;
             break;
         case 'h':
             default_wireless_hello_interval = parse_msec(optarg);
@@ -160,21 +158,21 @@ main(int argc, char **argv)
                 goto usage;
             break;
         case 'k':
-            kernel_metric = atoi(optarg);
+            kernel_metric = parse_nat(optarg);
             if(kernel_metric < 0 || kernel_metric > 0xFFFF)
                 goto usage;
             break;
         case 'A':
-            allow_duplicates = atoi(optarg);
+            allow_duplicates = parse_nat(optarg);
             if(allow_duplicates < 0 || allow_duplicates > 0xFFFF)
                 goto usage;
             break;
-        case 'P':
-            parasitic = 1;
-            break;
         case 's':
             split_horizon = 0;
             break;
+        case 'r':
+            random_id = 1;
+            break;
         case 'u':
             keep_unfeasible = 1;
             break;
@@ -182,13 +180,17 @@ main(int argc, char **argv)
             state_file = optarg;
             break;
         case 'd':
-            debug = atoi(optarg);
+            debug = parse_nat(optarg);
+            if(debug < 0)
+                goto usage;
             break;
         case 'g':
 #ifdef NO_LOCAL_INTERFACE
             fprintf(stderr, "Warning: no local interface in this version.\n");
 #else
-            local_server_port = atoi(optarg);
+            local_server_port = parse_nat(optarg);
+            if(local_server_port <= 0 || local_server_port > 0xFFFF)
+                goto usage;
 #endif
             break;
         case 'l':
@@ -199,24 +201,32 @@ main(int argc, char **argv)
             break;
         case 'z':
             {
-                char *comma = strchr(optarg, ',');
-                diversity_kind = atoi(optarg);
-                if(comma == NULL)
+                char *comma;
+                diversity_kind = (int)strtol(optarg, &comma, 0);
+                if(*comma == '\0')
                     diversity_factor = 128;
+                else if(*comma == ',')
+                    diversity_factor = parse_nat(comma + 1);
                 else
-                    diversity_factor = atoi(comma + 1);
+                    goto usage;
                 if(diversity_factor <= 0 || diversity_factor > 256)
                     goto usage;
             }
             break;
+        case 'M': {
+            int l = parse_nat(optarg);
+            if(l < 0 || l > 3600)
+                goto usage;
+            change_smoothing_half_life(l);
+            break;
+        }
         case 't':
-            export_table = atoi(optarg);
+            export_table = parse_nat(optarg);
             if(export_table < 0 || export_table > 0xFFFF)
                 goto usage;
             break;
         case 'T':
-            import_table = atoi(optarg);
-            if(import_table < 0 || import_table > 0xFFFF)
+            if(add_import_table(parse_nat(optarg)))
                 goto usage;
             break;
         case 'c':
@@ -249,11 +259,13 @@ main(int argc, char **argv)
             config_file = "/etc/babeld.conf";
     }
     if(config_file) {
-        rc = parse_config_from_file(config_file);
+        int line;
+        rc = parse_config_from_file(config_file, &line);
         if(rc < 0) {
             fprintf(stderr,
-                    "Couldn't parse configuration from file %s.\n",
-                    config_file);
+                    "Couldn't parse configuration from file %s "
+                    "(error at line %d).\n",
+                    config_file, line);
             exit(1);
         }
     } else {
@@ -275,12 +287,6 @@ main(int argc, char **argv)
     resend_delay = MIN(resend_delay, default_wired_hello_interval / 2);
     resend_delay = MAX(resend_delay, 20);
 
-    if(parasitic && allow_duplicates >= 0) {
-        /* Too difficult to get right. */
-        fprintf(stderr, "Sorry, -P and -A are incompatible.\n");
-        exit(1);
-    }
-
     if(do_daemonise) {
         if(logfile == NULL)
             logfile = "/var/log/babeld.log";
@@ -372,6 +378,9 @@ main(int argc, char **argv)
         goto fail;
     }
 
+    if(random_id)
+        goto random_id;
+
     FOR_ALL_INTERFACES(ifp) {
         /* ifp->ifindex is not necessarily valid at this point */
         int ifindex = if_nametoindex(ifp->name);
@@ -403,6 +412,7 @@ main(int argc, char **argv)
     fprintf(stderr,
             "Warning: couldn't find router id -- using random value.\n");
 
+ random_id:
     rc = read_random_bytes(myid, 8);
     if(rc < 0) {
         perror("read(random)");
@@ -412,7 +422,6 @@ main(int argc, char **argv)
     myid[0] &= ~3;
 
  have_id:
-    reboot_time = now.tv_sec;
     myseqno = (random() & 0xFFFF);
 
     fd = open(state_file, O_RDONLY);
@@ -448,11 +457,8 @@ main(int argc, char **argv)
                     gettimeofday(&realnow, NULL);
                     if(memcmp(sid, myid, 8) == 0)
                         myseqno = seqno_plus(s, 1);
-                    else
+                    else if(!random_id)
                         fprintf(stderr, "ID mismatch in babel-state.\n");
-                    /* Convert realtime into monotonic time. */
-                    if(t >= 1176800000L && t <= realnow.tv_sec)
-                        reboot_time = now.tv_sec - (realnow.tv_sec - t);
                 }
             } else {
                 fprintf(stderr, "Couldn't parse babel-state.\n");
@@ -482,18 +488,21 @@ main(int argc, char **argv)
     rc = resize_receive_buffer(1500);
     if(rc < 0)
         goto fail;
-    check_interfaces();
     if(receive_buffer == NULL)
         goto fail;
 
+    check_interfaces();
+
     rc = check_xroutes(0);
     if(rc < 0)
         fprintf(stderr, "Warning: couldn't check exported routes.\n");
+
     kernel_routes_changed = 0;
     kernel_link_changed = 0;
     kernel_addr_changed = 0;
     kernel_dump_time = now.tv_sec + roughly(30);
     schedule_neighbours_check(5000, 1);
+    schedule_interfaces_check(30000, 1);
     expiry_time = now.tv_sec + roughly(30);
     source_expiry_time = now.tv_sec + roughly(300);
 
@@ -531,6 +540,7 @@ main(int argc, char **argv)
         gettime(&now);
 
         tv = check_neighbours_timeout;
+        timeval_min(&tv, &check_interfaces_timeout);
         timeval_min_sec(&tv, expiry_time);
         timeval_min_sec(&tv, source_expiry_time);
         timeval_min_sec(&tv, kernel_dump_time);
@@ -556,13 +566,15 @@ main(int argc, char **argv)
                 maxfd = MAX(maxfd, kernel_socket);
             }
 #ifndef NO_LOCAL_INTERFACE
-            if(local_socket >= 0) {
-                FD_SET(local_socket, &readfds);
-                maxfd = MAX(maxfd, local_socket);
-            } else if(local_server_socket >= 0) {
+            if(local_server_socket >= 0 &&
+               num_local_sockets < MAX_LOCAL_SOCKETS) {
                 FD_SET(local_server_socket, &readfds);
                 maxfd = MAX(maxfd, local_server_socket);
             }
+            for(i = 0; i < num_local_sockets; i++) {
+                FD_SET(local_sockets[i], &readfds);
+                maxfd = MAX(maxfd, local_sockets[i]);
+            }
 #endif
             rc = select(maxfd + 1, &readfds, NULL, NULL, &tv);
             if(rc < 0) {
@@ -610,27 +622,38 @@ main(int argc, char **argv)
 #ifndef NO_LOCAL_INTERFACE
         if(local_server_socket >= 0 &&
            FD_ISSET(local_server_socket, &readfds)) {
-            if(local_socket >= 0) {
-                close(local_socket);
-                local_socket = -1;
-            }
-            local_socket = accept(local_server_socket, NULL, NULL);
-            if(local_socket < 0) {
+            int s;
+            s = accept(local_server_socket, NULL, NULL);
+            if(s < 0) {
                 if(errno != EINTR && errno != EAGAIN)
                     perror("accept(local_server_socket)");
+            } else if(num_local_sockets >= MAX_LOCAL_SOCKETS) {
+                /* This should never happen, since we don't select for
+                   the server socket in this case.  But I'm paranoid. */
+                fprintf(stderr, "Internal error: too many local sockets.\n");
+                close(s);
             } else {
-                local_notify_all();
+                local_sockets[num_local_sockets++] = s;
+                local_notify_all_1(s);
             }
         }
 
-        if(local_socket >= 0 && FD_ISSET(local_socket, &readfds)) {
-            rc = local_read(local_socket);
-            if(rc <= 0) {
-                if(rc < 0)
-                    perror("read(local_socket)");
-                close(local_socket);
-                local_socket = -1;
+        i = 0;
+        while(i < num_local_sockets) {
+            if(FD_ISSET(local_sockets[i], &readfds)) {
+                rc = local_read(local_sockets[i]);
+                if(rc <= 0) {
+                    if(rc < 0) {
+                        if(errno == EINTR)
+                            continue;
+                        perror("read(local_socket)");
+                    }
+                    close(local_sockets[i]);
+                    local_sockets[i] = local_sockets[--num_local_sockets];
+                    continue;
+                }
             }
+            i++;
         }
 #endif
 
@@ -666,12 +689,17 @@ main(int argc, char **argv)
         if(timeval_compare(&check_neighbours_timeout, &now) < 0) {
             int msecs;
             msecs = check_neighbours();
-            msecs = MAX(msecs, 10);
+            /* Multiply by 3/2 to allow neighbours to expire. */
+            msecs = MAX(3 * msecs / 2, 10);
             schedule_neighbours_check(msecs, 1);
         }
 
-        if(now.tv_sec >= expiry_time) {
+        if(timeval_compare(&check_interfaces_timeout, &now) < 0) {
             check_interfaces();
+            schedule_interfaces_check(30000, 1);
+        }
+
+        if(now.tv_sec >= expiry_time) {
             expire_routes();
             expire_resend();
             expiry_time = now.tv_sec + roughly(30);
@@ -786,7 +814,7 @@ main(int argc, char **argv)
             "                "
             "[-h hello] [-H wired_hello] [-z kind[,factor]]\n"
             "                "
-            "[-k metric] [-A metric] [-s] [-P] [-l] [-w] [-u] [-g port]\n"
+            "[-k metric] [-A metric] [-s] [-l] [-w] [-r] [-u] [-g port]\n"
             "                "
             "[-t table] [-T table] [-c file] [-C statement]\n"
             "                "
@@ -810,19 +838,30 @@ main(int argc, char **argv)
     exit(1);
 }
 
-/* Schedule a neighbours check after roughly 3/2 times msecs have elapsed. */
 void
 schedule_neighbours_check(int msecs, int override)
 {
     struct timeval timeout;
 
-    timeval_add_msec(&timeout, &now, roughly(msecs * 3 / 2));
+    timeval_add_msec(&timeout, &now, roughly(msecs));
     if(override)
         check_neighbours_timeout = timeout;
     else
         timeval_min(&check_neighbours_timeout, &timeout);
 }
 
+void
+schedule_interfaces_check(int msecs, int override)
+{
+    struct timeval timeout;
+
+    timeval_add_msec(&timeout, &now, roughly(msecs));
+    if(override)
+        check_interfaces_timeout = timeout;
+    else
+        timeval_min(&check_interfaces_timeout, &timeout);
+}
+
 int
 resize_receive_buffer(int size)
 {
@@ -938,7 +977,7 @@ dump_route_callback(struct babel_route *route, void *closure)
                 break;
             if(k > 0)
                 channels[j++] = ',';
-            snprintf(channels + j, 100 - j, "%d", route->channels[k]);
+            snprintf(channels + j, 100 - j, "%u", (unsigned)route->channels[k]);
             j = strlen(channels);
         }
         snprintf(channels + j, 100 - j, ")");
@@ -946,10 +985,10 @@ dump_route_callback(struct babel_route *route, void *closure)
             channels[0] = '\0';
     }
 
-    fprintf(out, "%s metric %d refmetric %d id %s seqno %d%s age %d "
+    fprintf(out, "%s metric %d (%d) refmetric %d id %s seqno %d%s age %d "
             "via %s neigh %s%s%s%s\n",
             format_prefix(route->src->prefix, route->src->plen),
-            route_metric(route), route->refmetric,
+            route_metric(route), route_smoothed_metric(route), route->refmetric,
             format_eui64(route->src->id),
             (int)route->seqno,
             channels,
