@@ -85,19 +85,21 @@ extern int debug;
 extern time_t reboot_time;
 extern int default_wireless_hello_interval, default_wired_hello_interval;
 extern int resend_delay;
+extern int do_daemonise;
+extern char *logfile, *pidfile, *state_file;
 extern int link_detect;
 extern int all_wireless;
-extern int local_socket;
 
 extern unsigned char myid[8];
 
 extern const unsigned char zeroes[16], ones[16];
 
-extern int protocol_port;
+extern int protocol_port, local_server_port;
 extern unsigned char protocol_group[16];
 extern int protocol_socket;
 extern int kernel_socket;
 extern int max_request_hopcount;
 
 void schedule_neighbours_check(int msecs, int override);
+void schedule_interfaces_check(int msecs, int override);
 int resize_receive_buffer(int size);
