@@ -35,13 +35,23 @@ THE SOFTWARE.
 #include "babeld.h"
 #include "util.h"
 #include "interface.h"
+#include "route.h"
+#include "kernel.h"
 #include "configuration.h"
 
 struct filter *input_filters = NULL;
 struct filter *output_filters = NULL;
 struct filter *redistribute_filters = NULL;
+struct interface_conf *default_interface_conf = NULL;
 struct interface_conf *interface_confs = NULL;
 
+/* This file implements a recursive descent parser with one character
+   lookahead.  The looked-ahead character is returned from most
+   functions.
+
+   Throughout this file, -1 signals that the look-ahead is EOF,
+   while -2 signals an error. */
+
 /* get_next_char callback */
 typedef int (*gnc_t)(void*);
 
@@ -70,15 +80,13 @@ getword(int c, char **token_r, gnc_t gnc, void *closure)
     int i = 0;
 
     c = skip_whitespace(c, gnc, closure);
-    if(c < 0)
-        return c;
-    if(c == '"' || c == '\n')
+    if(c < 0 || c == '"' || c == '\n' || c == '#' || c < 0)
         return -2;
     do {
         if(i >= 255) return -2;
         buf[i++] = c;
         c = gnc(closure);
-    } while(c != ' ' && c != '\t' && c != '\n' && c >= 0);
+    } while(c != ' ' && c != '\t' && c != '\n' && c != '#' && c >= 0);
     buf[i] = '\0';
     *token_r = strdup(buf);
     return c;
@@ -91,9 +99,7 @@ getstring(int c, char **token_r, gnc_t gnc, void *closure)
     int i = 0;
 
     c = skip_whitespace(c, gnc, closure);
-    if(c < 0)
-        return c;
-    if(c == '\n')
+    if(c < 0 || c == '\n' || c == '#')
         return -2;
 
     /* Unquoted strings have the same syntax as words. */
@@ -110,6 +116,10 @@ getstring(int c, char **token_r, gnc_t gnc, void *closure)
         }
         if(c == '\\')
             c = gnc(closure);
+
+        if(c < 0)
+            return -2;
+
         buf[i++] = c;
         c = gnc(closure);
     }
@@ -263,10 +273,9 @@ getnet(int c, unsigned char **p_r, unsigned char *plen_r, int *af_r,
     return c;
 }
 
-static struct filter *
-parse_filter(gnc_t gnc, void *closure)
+static int
+parse_filter(int c, gnc_t gnc, void *closure, struct filter **filter_return)
 {
-    int c;
     char *token;
     struct filter *filter;
 
@@ -275,13 +284,9 @@ parse_filter(gnc_t gnc, void *closure)
         goto error;
     filter->plen_le = 128;
 
-    c = gnc(closure);
-    if(c < -1)
-        goto error;
-
-    while(c >= 0 && c != '\n') {
+    while(1) {
         c = skip_whitespace(c, gnc, closure);
-        if(c == '\n' || c == '#') {
+        if(c < 0 || c == '\n' || c == '#') {
             c = skip_to_eol(c, gnc, closure);
             break;
         }
@@ -314,13 +319,13 @@ parse_filter(gnc_t gnc, void *closure)
                 goto error;
             filter->plen_ge = MAX(filter->plen_ge, p);
         } else if(strcmp(token, "neigh") == 0) {
-            unsigned char *neigh;
+            unsigned char *neigh = NULL;
             c = getip(c, &neigh, NULL, gnc, closure);
             if(c < -1)
                 goto error;
             filter->neigh = neigh;
         } else if(strcmp(token, "id") == 0) {
-            unsigned char *id;
+            unsigned char *id = NULL;
             c = getid(c, &id, gnc, closure);
             if(c < -1)
                 goto error;
@@ -363,41 +368,31 @@ parse_filter(gnc_t gnc, void *closure)
         filter->plen_le += 96;
         filter->plen_ge += 96;
     }
-    return filter;
+    *filter_return = filter;
+    return c;
 
  error:
     free(filter);
-    return NULL;
+    return -2;
 }
 
-static struct interface_conf *
-parse_ifconf(gnc_t gnc, void *closure)
+static int
+parse_anonymous_ifconf(int c, gnc_t gnc, void *closure,
+                       struct interface_conf *if_conf,
+                       struct interface_conf **if_conf_return)
 {
-    int c;
+
     char *token;
-    struct interface_conf *if_conf;
 
-    if_conf = calloc(1, sizeof(struct interface_conf));
-    if(if_conf == NULL)
-        goto error;
+    if(if_conf == NULL) {
+        if_conf = calloc(1, sizeof(struct interface_conf));
+        if(if_conf == NULL)
+            goto error;
+    }
 
-    c = gnc(closure);
-    if(c < -1)
-        goto error;
-
-    c = skip_whitespace(c, gnc, closure);
-    if(c < -1 || c == '\n' || c == '#')
-        goto error;
-
-    c = getstring(c, &token, gnc, closure);
-    if(c < -1 || token == NULL)
-        goto error;
-
-    if_conf->ifname = token;
-
-    while(c >= 0 && c != '\n') {
+    while(1) {
         c = skip_whitespace(c, gnc, closure);
-        if(c == '\n' || c == '#') {
+        if(c < 0 || c == '\n' || c == '#') {
             c = skip_to_eol(c, gnc, closure);
             break;
         }
@@ -466,7 +461,7 @@ parse_ifconf(gnc_t gnc, void *closure)
 
             free(t);
 
-            if((if_conf->channel < 1 || if_conf->channel > 254) &&
+            if((if_conf->channel < 1 || if_conf->channel > 255) &&
                if_conf->channel != IF_CHANNEL_NONINTERFERING)
                 goto error;
         } else {
@@ -475,11 +470,40 @@ parse_ifconf(gnc_t gnc, void *closure)
         free(token);
     }
 
-    return if_conf;
+    *if_conf_return = if_conf;
+    return c;
 
  error:
     free(if_conf);
-    return NULL;
+    return -2;
+}
+
+static int
+parse_ifconf(int c, gnc_t gnc, void *closure,
+             struct interface_conf **if_conf_return)
+{
+    char *token;
+    struct interface_conf *if_conf;
+
+    if_conf = calloc(1, sizeof(struct interface_conf));
+    if(if_conf == NULL)
+        goto error;
+
+    c = skip_whitespace(c, gnc, closure);
+    if(c < -1 || c == '\n' || c == '#')
+        goto error;
+
+    c = getstring(c, &token, gnc, closure);
+    if(c < -1 || token == NULL)
+        goto error;
+
+    if_conf->ifname = token;
+
+    return parse_anonymous_ifconf(c, gnc, closure, if_conf, if_conf_return);
+
+ error:
+    free(if_conf);
+    return -2;
 }
 
 static void
@@ -499,14 +523,17 @@ add_filter(struct filter *filter, struct filter **filters)
 }
 
 static void
-merge_ifconf(struct interface_conf *dest, struct interface_conf *src)
+merge_ifconf(struct interface_conf *dest,
+             const struct interface_conf *src1,
+             const struct interface_conf *src2)
 {
-    assert(strcmp(dest->ifname, src->ifname) == 0);
 
 #define MERGE(field)                            \
     do {                                        \
-        if(src->field)                          \
-            dest->field = src->field;           \
+        if(src1->field)                         \
+            dest->field = src1->field;          \
+        else                                    \
+            dest->field = src2->field;          \
     } while(0)
 
     MERGE(hello_interval);
@@ -515,6 +542,8 @@ merge_ifconf(struct interface_conf *dest, struct interface_conf *src)
     MERGE(wired);
     MERGE(split_horizon);
     MERGE(lq);
+    MERGE(faraway);
+    MERGE(channel);
 
 #undef MERGE
 }
@@ -530,7 +559,7 @@ add_ifconf(struct interface_conf *if_conf, struct interface_conf **if_confs)
         if_c = *if_confs;
         while(if_c->next) {
             if(strcmp(if_c->ifname, if_conf->ifname) == 0) {
-                merge_ifconf(if_c, if_conf);
+                merge_ifconf(if_c, if_conf, if_c);
                 free(if_conf);
                 return;
             }
@@ -541,6 +570,125 @@ add_ifconf(struct interface_conf *if_conf, struct interface_conf **if_confs)
     }
 }
 
+static int
+parse_option(int c, gnc_t gnc, void *closure, char *token)
+{
+    if(strcmp(token, "protocol-port") == 0 ||
+       strcmp(token, "kernel-priority") == 0 ||
+       strcmp(token, "allow-duplicates") == 0 ||
+#ifndef NO_LOCAL_INTERFACE
+       strcmp(token, "local-port") == 0 ||
+#endif
+       strcmp(token, "export-table") == 0 ||
+       strcmp(token, "import-table") == 0) {
+        int v;
+        c = getint(c, &v, gnc, closure);
+        if(c < -1 || v <= 0 || v >= 0xFFFF)
+            goto error;
+
+        if(strcmp(token, "protocol-port") == 0)
+            protocol_port = v;
+        else if(strcmp(token, "kernel-priority") == 0)
+            kernel_metric = v;
+        else if(strcmp(token, "allow_duplicates") == 0)
+            allow_duplicates = v;
+#ifndef NO_LOCAL_INTERFACE
+        else if(strcmp(token, "local-port") == 0)
+            local_server_port = v;
+#endif
+        else if(strcmp(token, "export-table") == 0)
+            export_table = v;
+        else if(strcmp(token, "import-table") == 0)
+            add_import_table(v);
+        else
+            abort();
+    } else if(strcmp(token, "keep-unfeasible") == 0 ||
+              strcmp(token, "link-detect") == 0 ||
+              strcmp(token, "daemonise") == 0) {
+        int b;
+        c = getbool(c, &b, gnc, closure);
+        if(c < -1)
+            goto error;
+        b = (b == CONFIG_YES);
+        if(strcmp(token, "keep-unfeasible") == 0)
+            keep_unfeasible = b;
+        else if(strcmp(token, "link-detect") == 0)
+            link_detect = b;
+        else if(strcmp(token, "daemonise") == 0)
+            do_daemonise = b;
+        else
+            abort();
+    } else if(strcmp(token, "protocol-group") == 0) {
+        unsigned char *group = NULL;
+        c = getip(c, &group, NULL, gnc, closure);
+        if(c < -1)
+            goto error;
+        memcpy(protocol_group, group, 16);
+        free(group);
+    } else if(strcmp(token, "state-file") == 0 ||
+              strcmp(token, "log-file") == 0 ||
+              strcmp(token, "pid-file") == 0) {
+        char *file;
+        c = getstring(c, &file, gnc, closure);
+        if(c < -1)
+            goto error;
+        if(strcmp(token, "state-file") == 0)
+            state_file = file;
+        else if(strcmp(token, "log-file") == 0)
+            logfile = file;
+        else if(strcmp(token, "pid-file") == 0)
+            pidfile = file;
+        else
+            abort();
+    } else if(strcmp(token, "debug") == 0) {
+        int d;
+        c = getint(c, &d, gnc, closure);
+        if(d < 0)
+            goto error;
+        debug = d;
+    } else if(strcmp(token, "diversity") == 0) {
+        int d;
+        c = skip_whitespace(c, gnc, closure);
+        if(c >= '0' && c <= '9') {
+            c = getint(c, &d, gnc, closure);
+            if(c < -1)
+                goto error;
+        } else {
+            int b;
+            c = getbool(c, &b, gnc, closure);
+            if(c < -1)
+                goto error;
+            d = (b == CONFIG_YES) ? 3 : 0;
+        }
+        diversity_kind = d;
+    } else if(strcmp(token, "diversity-factor") == 0) {
+        int f;
+        c = getint(c, &f, gnc, closure);
+        if(c < -1 || f < 0 || f > 256)
+            goto error;
+        diversity_factor = f;
+    } else if(strcmp(token, "smoothing-half-life") == 0) {
+        int h;
+        c = getint(c, &h, gnc, closure);
+        if(c < -1 || h < 0)
+            goto error;
+        change_smoothing_half_life(h);
+    } else {
+        goto error;
+    }
+
+    c = skip_whitespace(c, gnc, closure);
+    if(c < 0 || c == '\n' || c == '#') {
+        c = skip_to_eol(c, gnc, closure);
+        return c;
+    }
+
+    /* Fall through */
+ error:
+    return -2;
+
+}
+
 static int
 parse_config(gnc_t gnc, void *closure)
 {
@@ -548,10 +696,10 @@ parse_config(gnc_t gnc, void *closure)
     char *token;
 
     c = gnc(closure);
-    if(c < 2)
+    if(c < -1)
         return -1;
 
-    while(c >= 0) {
+    while(1) {
         c = skip_whitespace(c, gnc, closure);
         if(c == '\n' || c == '#') {
             c = skip_to_eol(c, gnc, closure);
@@ -565,47 +713,81 @@ parse_config(gnc_t gnc, void *closure)
 
         if(strcmp(token, "in") == 0) {
             struct filter *filter;
-            filter = parse_filter(gnc, closure);
-            if(filter == NULL)
+            c = parse_filter(c, gnc, closure, &filter);
+            if(c < -1)
                 return -1;
             add_filter(filter, &input_filters);
         } else if(strcmp(token, "out") == 0) {
             struct filter *filter;
-            filter = parse_filter(gnc, closure);
-            if(filter == NULL)
+            c = parse_filter(c, gnc, closure, &filter);
+            if(c < -1)
                 return -1;
             add_filter(filter, &output_filters);
         } else if(strcmp(token, "redistribute") == 0) {
             struct filter *filter;
-            filter = parse_filter(gnc, closure);
-            if(filter == NULL)
+            c = parse_filter(c, gnc, closure, &filter);
+            if(c < -1)
                 return -1;
             add_filter(filter, &redistribute_filters);
         } else if(strcmp(token, "interface") == 0) {
             struct interface_conf *if_conf;
-            if_conf = parse_ifconf(gnc, closure);
-            if(if_conf == NULL)
+            c = parse_ifconf(c, gnc, closure, &if_conf);
+            if(c < -1)
                 return -1;
             add_ifconf(if_conf, &interface_confs);
+        } else if(strcmp(token, "default") == 0) {
+            struct interface_conf *if_conf;
+            c = parse_anonymous_ifconf(c, gnc, closure, NULL, &if_conf);
+            if(c < -1)
+                return -1;
+            if(default_interface_conf == NULL)
+                default_interface_conf = if_conf;
+            else {
+                merge_ifconf(default_interface_conf,
+                             if_conf, default_interface_conf);
+                free(if_conf);
+            }
         } else {
-            return -1;
+            c = parse_option(c, gnc, closure, token);
+            if(c < -1)
+                return -1;
         }
         free(token);
     }
     return 1;
 }
 
-int
-parse_config_from_file(char *filename)
-{
+struct file_state {
     FILE *f;
+    int line;
+};
+
+static int
+gnc_file(struct file_state *s)
+{
+    int c;
+    c = fgetc(s->f);
+    if(c == '\n')
+       s->line++;
+    return c;
+}
+
+int
+parse_config_from_file(const char *filename, int *line_return)
+{
+    struct file_state s = { NULL, 1 };
     int rc;
 
-    f = fopen(filename, "r");
-    if(f == NULL)
+    s.f = fopen(filename, "r");
+    if(s.f == NULL) {
+        *line_return = 0;
         return -1;
-    rc = parse_config((gnc_t)fgetc, f);
-    fclose(f);
+    }
+
+    rc = parse_config((gnc_t)gnc_file, &s);
+    fclose(s.f);
+
+    *line_return = s.line;
     return rc;
 }
 
@@ -765,6 +947,8 @@ finalise_config()
         if_conf = interface_confs;
         interface_confs = interface_confs->next;
         if_conf->next = NULL;
+        if(default_interface_conf)
+            merge_ifconf(if_conf, if_conf, default_interface_conf);
         vrc = add_interface(if_conf->ifname, if_conf);
         if(vrc == NULL) {
             fprintf(stderr, "Couldn't add interface %s.\n", if_conf->ifname);
