@@ -51,13 +51,25 @@ THE SOFTWARE.
 #include "util.h"
 #include "interface.h"
 
-int export_table = -1, import_table = -1;
+#ifndef MAX_INTERFACES
+#define MAX_INTERFACES 20
+#endif
+
+int export_table = -1, import_tables[MAX_IMPORT_TABLES], import_table_count = 0;
 
 static int old_forwarding = -1;
 static int old_ipv4_forwarding = -1;
 static int old_accept_redirects = -1;
 static int old_rp_filter = -1;
 
+struct old_if {
+    char *ifname;
+    int rp_filter;
+};
+
+struct old_if old_if[MAX_INTERFACES];
+int num_old_if = 0;
+
 static int dgram_socket = -1;
 
 #ifndef ARPHRD_ETHER
@@ -468,8 +480,8 @@ kernel_setup(int setup)
         if(export_table < 0)
             export_table = RT_TABLE_MAIN;
 
-        if(import_table < 0)
-            import_table = RT_TABLE_MAIN;
+        if(import_table_count < 1)
+            import_tables[import_table_count++] = RT_TABLE_MAIN;
 
         dgram_socket = socket(PF_INET, SOCK_DGRAM, 0);
         if(dgram_socket < 0)
@@ -568,7 +580,7 @@ kernel_setup(int setup)
 
         if(old_rp_filter >= 0) {
             rc = write_proc("/proc/sys/net/ipv4/conf/all/rp_filter",
-                            old_accept_redirects);
+                            old_rp_filter);
             if(rc < 0) {
                 perror("Couldn't write rp_filter knob.\n");
                 return -1;
@@ -614,9 +626,58 @@ kernel_setup_socket(int setup)
     }
 }
 
+static int
+get_old_if(const char *ifname)
+{
+    int i;
+    for(i = 0; i < num_old_if; i++)
+        if(strcmp(old_if->ifname, ifname) == 0)
+            return i;
+    if(num_old_if >= MAX_INTERFACES)
+        return -1;
+    old_if[num_old_if].ifname = strdup(ifname);
+    if(old_if[num_old_if].ifname == NULL)
+        return -1;
+    old_if[num_old_if].rp_filter = -1;
+    return num_old_if++;
+}
+
 int
 kernel_setup_interface(int setup, const char *ifname, int ifindex)
 {
+    char buf[100];
+    int i, rc;
+
+    /* rp_filter has weird semantics: both all/rp_filter and ifname/rp_filter
+       must be set to 0 for the rp_filter to be disabled.  Deal with it. */
+
+    rc = snprintf(buf, 100, "/proc/sys/net/ipv4/conf/%s/rp_filter", ifname);
+    if(rc < 0 || rc >= 100)
+        return -1;
+
+    i = get_old_if(ifname);
+    if(setup) {
+        if(i >= 0)
+            old_if[i].rp_filter = read_proc(buf);
+        else
+            fprintf(stderr,
+                    "Warning: cannot save old configuration for %s.\n",
+                    ifname);
+        rc = write_proc(buf, 0);
+        if(rc < 0)
+            return -1;
+    } else {
+        if(i >= 0 && old_if[i].rp_filter >= 0)
+            rc = write_proc(buf, old_if[i].rp_filter);
+        else
+            rc = -1;
+
+        if(rc < 0)
+            fprintf(stderr,
+                    "Warning: cannot restore old configuration for %s.\n",
+                    ifname);
+    }
+
     return 1;
 }
 
@@ -712,6 +773,25 @@ isbridge(const char *ifname, int ifindex)
     return 0;
 }
 
+static int
+isbatman(const char *ifname, int ifindex)
+{
+    char buf[256];
+    int rc;
+
+    rc = snprintf(buf, 256, "/sys/devices/virtual/net/%s/mesh", ifname);
+    if(rc < 0 || rc >= 256)
+        return -1;
+
+    if(access(buf, F_OK) >= 0)
+        return 1;
+
+    if(errno != ENOENT)
+        return -1;
+
+    return 0;
+}
+
 int
 kernel_interface_wireless(const char *ifname, int ifindex)
 {
@@ -721,10 +801,8 @@ kernel_interface_wireless(const char *ifname, int ifindex)
     struct ifreq req;
     int rc;
 
-    if(isbridge(ifname, ifindex) != 0) {
-        /* Give up. */
+    if(isbridge(ifname, ifindex) != 0 || isbatman(ifname, ifindex) != 0)
         return -1;
-    }
 
     memset(&req, 0, sizeof(req));
     strncpy(req.ifr_name, ifname, sizeof(req.ifr_name));
@@ -864,43 +942,27 @@ kernel_route(int operation, const unsigned char *dest, unsigned short plen,
     ipv4 = v4mapped(gate);
 
     if(operation == ROUTE_MODIFY) {
-        int added;
         if(newmetric == metric && memcmp(newgate, gate, 16) == 0 &&
            newifindex == ifindex)
             return 0;
-        /* It is better to add the new route before removing the old
-           one, to avoid losing packets.  However, if the old and new
-           priorities are equal, this only works if the kernel supports
-           ECMP.  So we first try the "right" order, and fall back on
-           the "wrong" order if it fails with EEXIST. */
+        /* It would be better to add the new route before removing the
+           old one, to avoid losing packets.  However, this causes
+           problems with non-multipath kernels, which sometimes
+           silently fail the request, causing "stuck" routes.  Let's
+           stick with the naive approach, and hope that the window is
+           small enough to be negligible. */
+        kernel_route(ROUTE_FLUSH, dest, plen,
+                     gate, ifindex, metric,
+                     NULL, 0, 0);
         rc = kernel_route(ROUTE_ADD, dest, plen,
                           newgate, newifindex, newmetric,
                           NULL, 0, 0);
         if(rc < 0) {
-            if(errno != EEXIST)
-                return rc;
-            added = 0;
-        } else {
-            added = 1;
+            if(errno == EEXIST)
+                rc = 1;
+            /* Should we try to re-install the flushed route on failure?
+               Error handling is hard. */
         }
-
-        kernel_route(ROUTE_FLUSH, dest, plen,
-                     gate, ifindex, metric,
-                     NULL, 0, 0);
-
-        if(!added) {
-            rc = kernel_route(ROUTE_ADD, dest, plen,
-                              newgate, newifindex, newmetric,
-                              NULL, 0, 0);
-            if(rc < 0) {
-                if(errno == EEXIST)
-                    rc = 1;
-                /* In principle, we should try to re-install the flushed
-                   route on failure to preserve.  However, this should
-                   hopefully not matter much in practice. */
-            }
-        }
-
         return rc;
     }
 
@@ -1036,10 +1098,11 @@ parse_kernel_route_rta(struct rtmsg *rtm, int len, struct kernel_route *route)
     }
 #undef COPY_ADDR
 
-    if(table != import_table)
-        return -1;
-
-    return 0;
+    int i;
+    for(i = 0; i < import_table_count; i++)
+        if(table == import_tables[i])
+            return 0;
+    return -1;
 }
 
 static void
@@ -1105,6 +1168,10 @@ filter_kernel_routes(struct nlmsghdr *nh, void *data)
     if(rtm->rtm_src_len != 0)
         return 0;
 
+    /* Ignore cached routes, advertised by some kernels (linux 3.x). */
+    if(rtm->rtm_flags & RTM_F_CLONED)
+        return 0;
+
     if(data)
         current_route = &routes[*found];
     else
