@@ -57,7 +57,7 @@ static int get_sdl(struct sockaddr_dl *sdl, char *ifname);
 static const unsigned char v4prefix[16] =
     {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xFF, 0xFF, 0, 0, 0, 0 };
 
-int export_table = -1, import_table = -1;
+int export_table = -1, import_table_count = 0, import_tables[MAX_IMPORT_TABLES];
 
 int
 if_eui64(char *ifname, int ifindex, unsigned char *eui)
@@ -108,6 +108,8 @@ get_sdl(struct sockaddr_dl *sdl, char *ifname)
         return -1;
 
     buffer = (char *)malloc(buf_len);
+    if(buffer == NULL)
+        return -1;
 
     rc = sysctl(mib, 6, buffer, &buf_len, NULL, 0);
     if(rc < 0)
@@ -143,9 +145,13 @@ fail:
         (a).s6_addr[3] = (i) & 0xff;            \
     } while (0)
 
-#define ROUNDUP(a)                                                      \
+#if defined(__APPLE__)
+#define ROUNDUP(a) \
+    ((a) > 0 ? (1 + (((a) - 1) | (sizeof(uint32_t) - 1))) : sizeof(uint32_t))
+#else
+#define ROUNDUP(a) \
     ((a) > 0 ? (1 + (((a) - 1) | (sizeof(long) - 1))) : sizeof(long))
-
+#endif
 
 static int old_forwarding = -1;
 static int old_accept_redirects = -1;
@@ -154,17 +160,16 @@ static int ifindex_lo = -1;
 static int seq;
 
 static int
-mask2len(const struct in6_addr *addr)
+mask2len(const unsigned char *p, const int size)
 {
     int i = 0, j;
-    const u_char *p = (const u_char *)addr;
 
-    for(j = 0; j < 16; j++, p++) {
+    for(j = 0; j < size; j++, p++) {
         if(*p != 0xff)
             break;
         i += 8;
     }
-    if(j < 16) {
+    if(j < size) {
         switch(*p) {
 #define	MASKLEN(m, l)	case m: do { i += l; break; } while (0)
             MASKLEN(0xfe, 7); break;
@@ -232,7 +237,11 @@ kernel_setup(int setup)
 
     rc = 0;
     mib[2] = IPPROTO_ICMPV6;
+#if defined(IPV6CTL_SENDREDIRECTS)
+    mib[3] = IPV6CTL_SENDREDIRECTS;
+#else
     mib[3] = ICMPV6CTL_REDIRACCEPT;
+#endif
     datasize = sizeof(old_accept_redirects);
     if (setup)
         rc = sysctl(mib, 4, &old_accept_redirects, &datasize,
@@ -382,11 +391,12 @@ kernel_route(int operation, const unsigned char *dest, unsigned short plen,
              const unsigned char *newgate, int newifindex,
              unsigned int newmetric)
 {
-    unsigned char msg[512];
-    struct rt_msghdr *rtm;
-    struct sockaddr_in6 *sin6;
-    struct sockaddr_in *sin;
-    int rc, len, ipv4;
+    struct {
+      struct rt_msghdr m_rtm;
+      char m_space[512];
+    } msg;
+    char *data = msg.m_space;
+    int rc, ipv4;
 
     char local6[1][1][16] = IN6ADDR_LOOPBACK_INIT;
     char local4[1][1][16] =
@@ -413,9 +423,20 @@ kernel_route(int operation, const unsigned char *dest, unsigned short plen,
       return 0;
 
     if(operation == ROUTE_MODIFY) {
-        metric = newmetric;
-        gate = newgate;
-        ifindex = newifindex;
+        /* Do not use ROUTE_MODIFY when changing to a neighbour.
+           It is the only way to remove the "gateway" flag. */
+        if(ipv4 && plen == 128 && memcmp(dest, newgate, 16) == 0) {
+            kernel_route(ROUTE_FLUSH, dest, plen,
+                         gate, ifindex, metric,
+                         NULL, 0, 0);
+            return kernel_route(ROUTE_ADD, dest, plen,
+                                newgate, newifindex, newmetric,
+                                NULL, 0, 0);
+        } else {
+            metric = newmetric;
+            gate = newgate;
+            ifindex = newifindex;
+        }
     }
 
     kdebugf("kernel_route: %s %s/%d metric %d dev %d nexthop %s\n",
@@ -427,103 +448,107 @@ kernel_route(int operation, const unsigned char *dest, unsigned short plen,
     if(kernel_socket < 0) kernel_setup_socket(1);
 
     memset(&msg, 0, sizeof(msg));
-    rtm = (struct rt_msghdr *)msg;
-    rtm->rtm_version = RTM_VERSION;
+    msg.m_rtm.rtm_version = RTM_VERSION;
     switch(operation) {
     case ROUTE_FLUSH:
-        rtm->rtm_type = RTM_DELETE; break;
+        msg.m_rtm.rtm_type = RTM_DELETE; break;
     case ROUTE_ADD:
-        rtm->rtm_type = RTM_ADD; break;
-    case ROUTE_MODIFY: 
-        rtm->rtm_type = RTM_CHANGE; break;
-    default: 
+        msg.m_rtm.rtm_type = RTM_ADD; break;
+    case ROUTE_MODIFY:
+        msg.m_rtm.rtm_type = RTM_CHANGE; break;
+    default:
         return -1;
     };
-    rtm->rtm_index = ifindex;
-    rtm->rtm_flags = RTF_UP | RTF_PROTO2;
-    if(plen == 128) rtm->rtm_flags |= RTF_HOST;
-    /*     if(memcmp(nexthop->id, dest, 16) == 0) { */
-    /*         rtm -> rtm_flags |= RTF_LLINFO; */
-    /*         rtm -> rtm_flags |= RTF_CLONING; */
-    /*     } else { */
-    rtm->rtm_flags |= RTF_GATEWAY;
-    /*     } */
+    msg.m_rtm.rtm_index = ifindex;
+    msg.m_rtm.rtm_flags = RTF_UP;
+    if(plen == 128) msg.m_rtm.rtm_flags |= RTF_HOST;
     if(metric == KERNEL_INFINITY) {
-        rtm->rtm_flags |= RTF_BLACKHOLE;
+        msg.m_rtm.rtm_flags |= RTF_BLACKHOLE;
         if(ifindex_lo < 0) {
             ifindex_lo = if_nametoindex("lo0");
             if(ifindex_lo <= 0)
                 return -1;
         }
-        rtm->rtm_index = ifindex_lo;      
-    }
-    rtm->rtm_seq = ++seq;
-    rtm->rtm_addrs = RTA_DST | RTA_GATEWAY;
-    if(!(operation == ROUTE_MODIFY && plen == 128)) {
-        rtm->rtm_addrs |= RTA_NETMASK;
+        msg.m_rtm.rtm_index = ifindex_lo;
     }
+    msg.m_rtm.rtm_seq = ++seq;
+    msg.m_rtm.rtm_addrs = RTA_DST | RTA_GATEWAY;
+    if(plen != 128) msg.m_rtm.rtm_addrs |= RTA_NETMASK;
 
-#define push_sockaddr_in(ptr, offset) \
-    do { (ptr) = (struct sockaddr_in *)((char *)(ptr) + (offset)); \
-         (ptr)->sin_len = sizeof(struct sockaddr_in); \
-         (ptr)->sin_family = AF_INET; } while (0)
+#define PUSHEUI(ifindex) \
+    do { char ifname[IFNAMSIZ]; \
+         struct sockaddr_dl *sdl = (struct sockaddr_dl*) data; \
+         if(!if_indextoname((ifindex), ifname))  \
+             return -1; \
+         if(get_sdl(sdl, ifname) < 0)   \
+             return -1; \
+         data = data + ROUNDUP(sdl->sdl_len); \
+    } while (0)
 
-#define get_sin_addr(dst,src) \
-    do { memcpy((dst), (src) + 12, 4); } while (0)
+#define PUSHADDR(src) \
+    do { struct sockaddr_in *sin = (struct sockaddr_in*) data; \
+         sin->sin_len = sizeof(struct sockaddr_in);  \
+         sin->sin_family = AF_INET; \
+         memcpy(&sin->sin_addr, (src) + 12, 4); \
+         data = data + ROUNDUP(sin->sin_len); \
+    } while (0)
 
-#define push_sockaddr_in6(ptr, offset) \
-    do { (ptr) = (struct sockaddr_in6 *)((char *)(ptr) + (offset)); \
-         (ptr)->sin6_len = sizeof(struct sockaddr_in6); \
-         (ptr)->sin6_family = AF_INET6; } while (0)
-
-#define get_sin6_addr(dst,src) \
-    do { memcpy((dst), (src), 16); } while (0)
+#define PUSHADDR6(src) \
+    do { struct sockaddr_in6 *sin6 = (struct sockaddr_in6*) data; \
+         sin6->sin6_len = sizeof(struct sockaddr_in6); \
+         sin6->sin6_family = AF_INET6; \
+         memcpy(&sin6->sin6_addr, (src), 16); \
+         if(IN6_IS_ADDR_LINKLOCAL (&sin6->sin6_addr)) \
+            SET_IN6_LINKLOCAL_IFINDEX (sin6->sin6_addr, ifindex); \
+         data = data + ROUNDUP(sin6->sin6_len); \
+    } while (0)
 
     /* KAME ipv6 stack does not support IPv4 mapped IPv6, so we have to
      * duplicate the codepath */
     if(ipv4) {
-        sin = (struct sockaddr_in *)msg;
-        /* destination */
-        push_sockaddr_in(sin, sizeof(*rtm));
-        get_sin_addr(&(sin->sin_addr), dest);
-        /* gateway */
-        push_sockaddr_in(sin, ROUNDUP(sin->sin_len));
-        if (metric == KERNEL_INFINITY)
-            get_sin_addr(&(sin->sin_addr),**local4);
-        else
-            get_sin_addr(&(sin->sin_addr),gate);
-        /* netmask */
-        if((rtm->rtm_addrs | RTA_NETMASK) != 0) {
-            struct in6_addr tmp_sin6_addr;
-            push_sockaddr_in(sin, ROUNDUP(sin->sin_len));
-            plen2mask(plen, &tmp_sin6_addr);
-            get_sin_addr(&(sin->sin_addr), (char *)&tmp_sin6_addr);
-        }
-        len = (char *)sin + ROUNDUP(sin->sin_len) - (char *)msg;
-    } else {
-        sin6 = (struct sockaddr_in6 *)msg;
-        /* destination */
-        push_sockaddr_in6(sin6, sizeof(*rtm));
-        get_sin6_addr(&(sin6->sin6_addr), dest);
-        /* gateway */
-        push_sockaddr_in6(sin6, ROUNDUP(sin6->sin6_len));
-        if (metric == KERNEL_INFINITY)
-            get_sin6_addr(&(sin6->sin6_addr),**local6);
-        else
-            get_sin6_addr(&(sin6->sin6_addr),gate);
-        if(IN6_IS_ADDR_LINKLOCAL (&sin6->sin6_addr))
-            SET_IN6_LINKLOCAL_IFINDEX (sin6->sin6_addr, ifindex);
-        /* netmask */
-        if((rtm->rtm_addrs | RTA_NETMASK) != 0) {
-            push_sockaddr_in6(sin6, ROUNDUP(sin6->sin6_len));
-            plen2mask(plen, &sin6->sin6_addr);
-        }
-        len = (char *)sin6 + ROUNDUP(sin6->sin6_len) - (char *)msg;
-    }
-    rtm->rtm_msglen = len;
 
-    rc = write(kernel_socket, msg, rtm->rtm_msglen);
-    if (rc < rtm->rtm_msglen)
+        PUSHADDR(dest);
+        if (metric == KERNEL_INFINITY) {
+            PUSHADDR(**local4);
+        } else if(plen == 128 && memcmp(dest+12, gate+12, 4) == 0) {
+#if defined(RTF_CLONING)
+            msg.m_rtm.rtm_flags |= RTF_CLONING;
+#endif
+            PUSHEUI(ifindex);
+        } else {
+            msg.m_rtm.rtm_flags |= RTF_GATEWAY;
+            PUSHADDR(gate);
+        }
+        if((msg.m_rtm.rtm_addrs & RTA_NETMASK) != 0) {
+            struct in6_addr tmp_sin6_addr;
+            plen2mask(plen, &tmp_sin6_addr);
+            PUSHADDR((char *)&tmp_sin6_addr);
+        }
+
+    } else {
+
+        PUSHADDR6(dest);
+        if (metric == KERNEL_INFINITY) {
+            PUSHADDR6(**local6);
+        } else {
+            msg.m_rtm.rtm_flags |= RTF_GATEWAY;
+            PUSHADDR6(gate);
+        }
+        if((msg.m_rtm.rtm_addrs & RTA_NETMASK) != 0) {
+            struct in6_addr tmp_sin6_addr;
+            plen2mask(plen, &tmp_sin6_addr);
+            PUSHADDR6((char*)&tmp_sin6_addr);
+        }
+
+    }
+
+#undef PUSHEUI
+#undef PUSHADDR
+#undef PUSHADDR6
+
+    msg.m_rtm.rtm_msglen = data - (char *)&msg;
+    rc = write(kernel_socket, (char*)&msg, msg.m_rtm.rtm_msglen);
+    if (rc < msg.m_rtm.rtm_msglen)
         return -1;
 
     return 1;
@@ -533,70 +558,27 @@ static void
 print_kernel_route(int add, struct kernel_route *route)
 {
     char ifname[IFNAMSIZ];
-    char addr_prefix[INET6_ADDRSTRLEN];
-    char addr_gw[INET6_ADDRSTRLEN];
-    
-    if(!inet_ntop(AF_INET6, route->prefix,
-                  addr_prefix, sizeof(addr_prefix)) ||
-       !inet_ntop(AF_INET6,route->gw, addr_gw, sizeof(addr_gw)) ||
-       !if_indextoname(route->ifindex, ifname)) {
-        fprintf(stderr,"Couldn't format kernel route for printing.");
-        // return;
-    }
+
+    if(!if_indextoname(route->ifindex, ifname))
+      memcpy(ifname,"unk",4);
 
     fprintf(stderr,
-            "%s kernel route: dest: %s/%d gw: %s metric: %d if: %s(%d) \n",
+            "%s kernel route: dest: %s gw: %s metric: %d if: %s(%d) \n",
             add == RTM_ADD ? "Add" :
             add == RTM_DELETE ? "Delete" : "Change",
-            addr_prefix, route->plen, addr_gw, route->metric, ifname, 
-            route->ifindex
+            format_prefix(route->prefix, route->plen),
+            format_address(route->gw),
+            route->metric,
+            ifname, route->ifindex
             );
 }
 
 static int
 parse_kernel_route(const struct rt_msghdr *rtm, struct kernel_route *route)
 {
-
+    struct sockaddr *sa;
     void *rta = (void*)rtm + sizeof(struct rt_msghdr);
-    struct sockaddr_in6 *sin6;
-    char addr[INET6_ADDRSTRLEN];
-
-    memset(route, 0, sizeof(*route));
-    route->metric = 0;
-    route->ifindex = rtm->rtm_index;
-
-    if(!(rtm->rtm_addrs & RTA_DST))
-        return -1;
-    sin6 = (struct sockaddr_in6 *)rta;
-    if(IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr) 
-       || IN6_IS_ADDR_MC_LINKLOCAL(&sin6->sin6_addr))
-        return -1;
-    if((rtm->rtm_flags & RTF_PROTO2) != 0)
-        return -1;
-    memcpy(&route->prefix, &sin6->sin6_addr, 16);
-    rta += ROUNDUP(sizeof(struct sockaddr_in6));
-   
-    if(!(rtm->rtm_addrs & RTA_GATEWAY))
-        return -1;
-
-    sin6 = (struct sockaddr_in6 *)rta;
-    if(IN6_IS_ADDR_LINKLOCAL (&sin6->sin6_addr)) {
-        route->ifindex = IN6_LINKLOCAL_IFINDEX(sin6->sin6_addr);
-        SET_IN6_LINKLOCAL_IFINDEX(sin6->sin6_addr, 0);
-    }
-    memcpy(&route->gw, &sin6->sin6_addr, 16);
-    rta += ROUNDUP(sizeof(struct sockaddr_in6));
-
-    if(!(rtm->rtm_addrs & RTA_NETMASK)) {
-        route->plen = 0;
-    } else {
-        sin6 = (struct sockaddr_in6 *)rta;        
-        route->plen = mask2len(&sin6->sin6_addr);
-        inet_ntop(AF_INET6, &sin6->sin6_addr, addr, sizeof(addr));
-        rta += ROUNDUP(sizeof(struct sockaddr_in6));
-    }
-    if (rtm->rtm_flags & RTF_HOST)
-        route->plen = 128;
+    uint32_t excluded_flags = 0;
 
     if(ifindex_lo < 0) {
         ifindex_lo = if_nametoindex("lo0");
@@ -604,11 +586,78 @@ parse_kernel_route(const struct rt_msghdr *rtm, struct kernel_route *route)
             return -1;
     }
 
+    memset(route, 0, sizeof(*route));
+    route->metric = 0;
+    route->ifindex = rtm->rtm_index;
+
+#if defined(RTF_IFSCOPE)
+    /* Filter out kernel route on OS X */
+    excluded_flags |= RTF_IFSCOPE;
+#endif
+#if defined(RTF_MULTICAST)
+    /* Filter out multicast route on others BSD */
+    excluded_flags |= RTF_MULTICAST;
+#endif
+
+    /* Prefix */
+    if(!(rtm->rtm_addrs & RTA_DST))
+        return -1;
+    sa = (struct sockaddr *)rta;
+    rta += ROUNDUP(sa->sa_len);
+    if(sa->sa_family == AF_INET6) {
+          struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
+        memcpy(route->prefix, &sin6->sin6_addr, 16);
+        if(IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr)
+             || IN6_IS_ADDR_MC_LINKLOCAL(&sin6->sin6_addr))
+           return -1;
+    } else if(sa->sa_family == AF_INET) {
+        struct sockaddr_in *sin = (struct sockaddr_in *)sa;
+#if defined(IN_LINKLOCAL)
+        if(IN_LINKLOCAL(ntohl(sin->sin_addr.s_addr)))
+            return -1;
+#endif
+        if(IN_MULTICAST(ntohl(sin->sin_addr.s_addr)))
+            return -1;
+        v4tov6(route->prefix, (unsigned char *)&sin->sin_addr);
+    } else {
+        return -1;
+    }
+
+    /* Gateway */
+    if(!(rtm->rtm_addrs & RTA_GATEWAY))
+        return -1;
+    sa = (struct sockaddr *)rta;
+    rta += ROUNDUP(sa->sa_len);
+    if(sa->sa_family == AF_INET6) {
+          struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
+        memcpy(route->gw, &sin6->sin6_addr, 16);
+        if(IN6_IS_ADDR_LINKLOCAL (&sin6->sin6_addr)) {
+            route->ifindex = IN6_LINKLOCAL_IFINDEX(sin6->sin6_addr);
+            SET_IN6_LINKLOCAL_IFINDEX(sin6->sin6_addr, 0);
+        }
+    } else if(sa->sa_family == AF_INET) {
+        struct sockaddr_in *sin = (struct sockaddr_in *)sa;
+        v4tov6(route->gw, (unsigned char *)&sin->sin_addr);
+    }
     if(route->ifindex == ifindex_lo)
         return -1;
 
-    return 0;
+    /* Netmask */
+    if((rtm->rtm_addrs & RTA_NETMASK) != 0) {
+        sa = (struct sockaddr *)rta;
+        rta += ROUNDUP(sa->sa_len);
+        if(!v4mapped(route->prefix)) {
+            struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
+            route->plen = mask2len((unsigned char*)&sin6->sin6_addr, 16);
+        } else {
+            struct sockaddr_in *sin = (struct sockaddr_in *)sa;
+            route->plen = mask2len((unsigned char*)&sin->sin_addr, 4);
+        }
+    }
+    if(v4mapped(route->prefix)) route->plen += 96;
+    if(rtm->rtm_flags & RTF_HOST) route->plen = 128;
 
+    return 0;
 }
 
 int
@@ -623,7 +672,7 @@ kernel_routes(struct kernel_route *routes, int maxroutes)
     mib[0] = CTL_NET;
     mib[1] = PF_ROUTE;
     mib[2] = 0;
-    mib[3] = AF_INET6;	  /* Address family */
+    mib[3] = AF_UNSPEC;      /* Address family */
     mib[4] = NET_RT_DUMP; /* Dump the kernel routing table */
     mib[5] = 0;		  /* No flags */
 
@@ -749,6 +798,10 @@ kernel_addresses(char *ifname, int ifindex, int ll,
             struct sockaddr_in *sin = (struct sockaddr_in*)ifap->ifa_addr;
             if(ll)
                 goto next;
+#if defined(IN_LINKLOCAL)
+            if(IN_LINKLOCAL(htonl(sin->sin_addr.s_addr)))
+                goto next;
+#endif
             memcpy(routes[i].prefix, v4prefix, 12);
             memcpy(routes[i].prefix + 12, &sin->sin_addr, 4);
             routes[i].plen = 128;
